@@ -5,12 +5,11 @@ use strict;
 use base 'MogileFS::Worker';
 use fields (
             'fidtodo',   # hashref { fid => 1 }
-            'peerrepl',  # hashref { fid => time() } # when peer started replicating
-            'rebal_pol_obj',    # rebalancer policy object
             );
 
 use List::Util ();
 use MogileFS::Util qw(error every debug);
+use MogileFS::Config;
 use MogileFS::Class;
 use MogileFS::RebalancePolicy::DrainDevices;
 use MogileFS::ReplicationRequest qw(rr_upgrade);
@@ -26,7 +25,6 @@ sub new {
     my $self = fields::new($class);
     $self->SUPER::new($psock);
     $self->{fidtodo} = {};
-    $self->{peerrepl} = {};
     return $self;
 }
 
@@ -39,7 +37,7 @@ sub work {
     # give the monitor job 15 seconds to give us an update
     my $warn_after = time() + 15;
 
-    every(2.0, sub {
+    every(1.0, sub {
         # replication doesn't go well if the monitor job hasn't actively started
         # marking things as being available
         unless ($self->monitor_has_run) {
@@ -50,17 +48,34 @@ sub work {
 
         $self->validate_dbh;
         my $dbh = $self->get_dbh or return 0;
+        my $sto = Mgd::get_store();
+        $self->send_to_parent("worker_bored 100 replicate rebalance");
 
-        $self->send_to_parent("worker_bored 100 replicate drain rebalance");
-        $self->read_from_parent(1);
-        # this finds stuff to replicate based on its record in the needs_replication table
-        $self->replicate_using_torepl_table;
+        my $queue_todo  = $self->queue_todo('replicate');
+        my $queue_todo2 = $self->queue_todo('rebalance');
+        unless (@$queue_todo || @$queue_todo2) {
+            return;
+        }
 
-        # if replicators are otherwise idle, use them to make the world
-        # better, rebalancing things (if enabled), and draining devices (if
-        # any are marked drain)
-        $self->rebalance_devices;
-        $self->drain_devices;
+        while (my $todo = shift @$queue_todo) {
+            my $fid = $todo->{fid};
+            $self->replicate_using_torepl_table($todo);
+        }
+        while (my $todo = shift @$queue_todo2) {
+            $self->still_alive;
+            # deserialize the arg :/
+            $todo->{arg} = [split /,/, $todo->{arg}];
+            my $devfid =
+                MogileFS::DevFID->new($todo->{devid}, $todo->{fid});
+            $self->rebalance_devfid($devfid, 
+                { target_devids => $todo->{arg} });
+
+            # If files error out, we want to send the error up to syslog
+            # and make a real effort to chew through the queue. Users may
+            # manually re-run rebalance to retry.
+            $sto->delete_fid_from_file_to_queue($todo->{fid}, REBAL_QUEUE);
+        }
+        $_[0]->(0); # don't sleep.
     });
 }
 
@@ -68,185 +83,129 @@ sub work {
 # there was nothing to be done.
 sub replicate_using_torepl_table {
     my $self = shift;
+    my $todo = shift;
 
     # find some fids to replicate, prioritize based on when they should be tried
     my $sto = Mgd::get_store();
-    my $queue_todo = $self->queue_todo('replicate');
-    unless (@$queue_todo) {
-        $self->parent_ping;
-        return 0;
-    }
 
-    while (my $todo = shift @$queue_todo) {
-        next unless $todo->{_type} eq 'replicate';
-        my $fid = $todo->{fid};
-        $self->still_alive;
+    my $fid = $todo->{fid};
+    $self->still_alive;
 
-        my $errcode;
+    my $errcode;
 
-        my %opts;
-        $opts{errref}       = \$errcode;
-        $opts{no_unlock}    = 1; # to make it return an $unlock subref
-        $opts{source_devid} = $todo->{fromdevid} if $todo->{fromdevid};
+    my %opts;
+    $opts{errref}       = \$errcode;
+    $opts{no_unlock}    = 1; # to make it return an $unlock subref
+    $opts{source_devid} = $todo->{fromdevid} if $todo->{fromdevid};
 
-        my ($status, $unlock) = replicate($fid, %opts);
+    my ($status, $unlock) = replicate($fid, %opts);
 
-        if ($status) {
-            # $status is either 0 (failure, handled below), 1 (success, we actually
-            # replicated this file), or 2 (success, but someone else replicated it).
+    if ($status) {
+        # $status is either 0 (failure, handled below), 1 (success, we actually
+        # replicated this file), or 2 (success, but someone else replicated it).
 
-            # when $staus eq "lost_race", this delete is unnecessary normally
-            # (somebody else presumably already deleted it if they
-            # also replicated it), but in the case of running with old
-            # replicators from previous versions, -or- simply if the
-            # other guy's delete failed, this cleans it up....
-            $sto->delete_fid_from_file_to_replicate($fid);
-            $unlock->() if $unlock;
-            next;
-        }
-
-        debug("Replication of fid=$fid failed with errcode=$errcode") if $Mgd::DEBUG >= 2;
-
-        # ERROR CASES:
-
-        # README: please keep this up to date if you update the replicate() function so we ensure
-        # that this code always does the right thing
-        #
-        # -- HARMLESS --
-        # failed_getting_lock        => harmless.  skip.  somebody else probably doing.
-        #
-        # -- ACTIONABLE --
-        # too_happy                  => too many copies, attempt to rebalance.
-        #
-        # -- TEMPORARY; DO EXPONENTIAL BACKOFF --
-        # source_down                => only source available is observed down.
-        # policy_error_doing_failed  => policy plugin fucked up.  it's looping.
-        # policy_error_already_there => policy plugin fucked up.  it's dumb.
-        # policy_no_suggestions      => no copy was attempted.  policy is just not happy.
-        # copy_error                 => policy said to do 1+ things, we failed, it ran out of suggestions.
-        #
-        # -- FATAL; DON'T TRY AGAIN --
-        # no_source                  => it simply exists nowhere.  not that something's down, but file_on is empty.
-
-        # bail if we failed getting the lock, that means someone else probably
-        # already did it, so we should just move on
-        if ($errcode eq 'failed_getting_lock') {
-            $unlock->() if $unlock;
-            next;
-        }
-
-        # logic for setting the next try time appropriately
-        my $update_nexttry = sub {
-            my ($type, $delay) = @_;
-            my $sto = Mgd::get_store();
-            if ($type eq 'end_of_time') {
-                # special; update to a time that won't happen again,
-                # as we've encountered a scenario in which case we're
-                # really hosed
-                $sto->reschedule_file_to_replicate_absolute($fid, ENDOFTIME);
-            } elsif ($type eq "offset") {
-                $sto->reschedule_file_to_replicate_relative($fid, $delay+0);
-            } else {
-                $sto->reschedule_file_to_replicate_absolute($fid, $delay+0);
-            }
-        };
-
-        # now let's handle any error we want to consider a total failure; do not
-        # retry at any point.  push this file off to the end so someone has to come
-        # along and figure out what went wrong.
-        if ($errcode eq 'no_source') {
-            $update_nexttry->( end_of_time => 1 );
-            $unlock->() if $unlock;
-            next;
-        }
-
-        # try to shake off extra copies. fall through to the backoff logic
-        # so we don't flood if it's impossible to properly weaken the fid.
-        # there's a race where the fid could be checked again, but the
-        # exclusive locking prevents replication clobbering.
-        if ($errcode eq 'too_happy') {
-            $unlock->() if $unlock;
-            $unlock = undef;
-            my $f = MogileFS::FID->new($fid);
-            my @devs = List::Util::shuffle($f->devids);
-            my $devfid;
-            # First one we can delete from, we try to rebalance away from.
-            for (@devs) {
-                my $dev = MogileFS::Device->of_devid($_);
-                # Not positive 'can_read_from' needs to be here.
-                # We must be able to delete off of this dev so the fid can
-                # move.
-                if ($dev->can_delete_from && $dev->can_read_from) {
-                    $devfid = MogileFS::DevFID->new($dev, $f);
-                    last;
-                }
-            }
-            $self->rebalance_devfid($devfid) if $devfid;
-        }
-
-        # at this point, the rest of the errors require exponential backoff.  define what this means
-        # as far as failcount -> delay to next try.
-        # 15s, 1m, 5m, 30m, 1h, 2h, 4h, 8h, 24h, 24h, 24h, 24h, ...
-        my @backoff = qw( 15 60 300 1800 3600 7200 14400 28800 );
-        $update_nexttry->( offset => int(($backoff[$todo->{failcount}] || 86400) * (rand(0.4) + 0.8)) );
+        # when $staus eq "lost_race", this delete is unnecessary normally
+        # (somebody else presumably already deleted it if they
+        # also replicated it), but in the case of running with old
+        # replicators from previous versions, -or- simply if the
+        # other guy's delete failed, this cleans it up....
+        $sto->delete_fid_from_file_to_replicate($fid);
         $unlock->() if $unlock;
+        next;
     }
+
+    debug("Replication of fid=$fid failed with errcode=$errcode") if $Mgd::DEBUG >= 2;
+
+    # ERROR CASES:
+
+    # README: please keep this up to date if you update the replicate() function so we ensure
+    # that this code always does the right thing
+    #
+    # -- HARMLESS --
+    # failed_getting_lock        => harmless.  skip.  somebody else probably doing.
+    #
+    # -- ACTIONABLE --
+    # too_happy                  => too many copies, attempt to rebalance.
+    #
+    # -- TEMPORARY; DO EXPONENTIAL BACKOFF --
+    # source_down                => only source available is observed down.
+    # policy_error_doing_failed  => policy plugin fucked up.  it's looping.
+    # policy_error_already_there => policy plugin fucked up.  it's dumb.
+    # policy_no_suggestions      => no copy was attempted.  policy is just not happy.
+    # copy_error                 => policy said to do 1+ things, we failed, it ran out of suggestions.
+    #
+    # -- FATAL; DON'T TRY AGAIN --
+    # no_source                  => it simply exists nowhere.  not that something's down, but file_on is empty.
+
+    # bail if we failed getting the lock, that means someone else probably
+    # already did it, so we should just move on
+    if ($errcode eq 'failed_getting_lock') {
+        $unlock->() if $unlock;
+        next;
+    }
+
+    # logic for setting the next try time appropriately
+    my $update_nexttry = sub {
+        my ($type, $delay) = @_;
+        my $sto = Mgd::get_store();
+        if ($type eq 'end_of_time') {
+            # special; update to a time that won't happen again,
+            # as we've encountered a scenario in which case we're
+            # really hosed
+            $sto->reschedule_file_to_replicate_absolute($fid, ENDOFTIME);
+        } elsif ($type eq "offset") {
+            $sto->reschedule_file_to_replicate_relative($fid, $delay+0);
+        } else {
+            $sto->reschedule_file_to_replicate_absolute($fid, $delay+0);
+        }
+    };
+
+    # now let's handle any error we want to consider a total failure; do not
+    # retry at any point.  push this file off to the end so someone has to come
+    # along and figure out what went wrong.
+    if ($errcode eq 'no_source') {
+        $update_nexttry->( end_of_time => 1 );
+        $unlock->() if $unlock;
+        next;
+    }
+
+    # try to shake off extra copies. fall through to the backoff logic
+    # so we don't flood if it's impossible to properly weaken the fid.
+    # there's a race where the fid could be checked again, but the
+    # exclusive locking prevents replication clobbering.
+    if ($errcode eq 'too_happy') {
+        $unlock->() if $unlock;
+        $unlock = undef;
+        my $f = MogileFS::FID->new($fid);
+        my @devs = List::Util::shuffle($f->devids);
+        my $devfid;
+        # First one we can delete from, we try to rebalance away from.
+        for (@devs) {
+            my $dev = MogileFS::Device->of_devid($_);
+            # Not positive 'can_read_from' needs to be here.
+            # We must be able to delete off of this dev so the fid can
+            # move.
+            if ($dev->can_delete_from && $dev->can_read_from) {
+                $devfid = MogileFS::DevFID->new($dev, $f);
+                last;
+            }
+        }
+        $self->rebalance_devfid($devfid) if $devfid;
+    }
+
+    # at this point, the rest of the errors require exponential backoff.  define what this means
+    # as far as failcount -> delay to next try.
+    # 15s, 1m, 5m, 30m, 1h, 2h, 4h, 8h, 24h, 24h, 24h, 24h, ...
+    my @backoff = qw( 15 60 300 1800 3600 7200 14400 28800 );
+    $update_nexttry->( offset => int(($backoff[$todo->{failcount}] || 86400) * (rand(0.4) + 0.8)) );
+    $unlock->() if $unlock;
     return 1;
 }
 
-sub rebalance_devices {
-    my $self = shift;
-    my $sto = Mgd::get_store();
-    return 0 unless $sto->server_setting('enable_rebalance');
-    my $pol = $self->rebalance_policy_obj or return 0;
-    unless ($self->run_rebalance_policy_a_bit($pol)) {
-        error("disabling rebalancing due to lack of work");
-        MogileFS::Config->set_server_setting("enable_rebalance", 0);
-    }
-}
-
-sub drain_devices {
-    my $self = shift;
-    my $pol = MogileFS::RebalancePolicy::DrainDevices->instance;
-    my $rv = $self->run_rebalance_policy_a_bit($pol);
-    #error("[$$] drained = $rv\n") if $rv;
-}
-
-# returns number of files rebalanced.
-sub run_rebalance_policy_a_bit {
-    my ($self, $pol) = @_;
-    my $stop_at = time() + 5; # Run for up to 5 seconds, then return.
-    my $n = 0;
-    my %avoid_devids = map { $_->id => 1 } $pol->dest_devs_to_avoid;
-    while (my $dfid = $pol->devfid_to_rebalance) {
-        $self->rebalance_devfid($dfid, avoid_devids => \%avoid_devids);
-        $n++;
-        last if time() >= $stop_at;
-    }
-    return $n;
-}
-
-sub rebalance_policy_obj {
-    my $self = shift;
-    my $rclass = Mgd::get_store()->server_setting('rebalance_policy') ||
-        "MogileFS::RebalancePolicy::PercentFree";
-
-    # return old one, if it's still of the same type.
-    if ($self->{'rebal_pol_obj'} && ref($self->{'rebal_pol_obj'}) eq $rclass) {
-        return $self->{'rebal_pol_obj'};
-    }
-
-    return error("Bogus rebalance_policy setting") unless $rclass =~ /^[\w:\-]+$/;
-    return error("Failed to load $rclass: $@") unless eval "use $rclass; 1;";
-    my $pol = eval { $rclass->new };
-    return error("Failed to instantiate rebalance policy: $@") unless $pol;
-    return $self->{'rebal_pol_obj'} = $pol;
-}
-
 # Return 1 on success, 0 on failure.
 sub rebalance_devfid {
-    my ($self, $devfid, %opts) = @_;
-    MogileFS::Util::okay_args(\%opts, qw(avoid_devids));
+    my ($self, $devfid, $opts) = @_;
+    MogileFS::Util::okay_args($opts, qw(avoid_devids target_devids));
 
     my $fid = $devfid->fid;
 
@@ -262,7 +221,7 @@ sub rebalance_devfid {
     my ($ret, $unlock) = replicate($fid,
                                    mask_devids  => { $devfid->devid => 1 },
                                    no_unlock    => 1,
-                                   avoid_devids => $opts{avoid_devids},
+                                   target_devids => $opts->{target_devids},
                                    errref       => \$errcode,
                                    );
 
@@ -359,6 +318,7 @@ sub replicate {
     my $sdevid    = delete $opts{'source_devid'};
     my $mask_devids  = delete $opts{'mask_devids'}  || {};
     my $avoid_devids = delete $opts{'avoid_devids'} || {};
+    my $target_devids = delete $opts{'target_devids'}; # inverse of avoid_devids.
     die "unknown_opts" if %opts;
     die unless ref $mask_devids eq "HASH";
 
@@ -443,12 +403,17 @@ sub replicate {
     my $got_copy_request = 0;  # true once replication policy asks us to move something somewhere
     my $copy_err;
 
+    my $dest_devs = $devs;
+    if ($target_devids) {
+        $dest_devs = {map { $_ => $devs->{$_} } @$target_devids};
+    }
+
     my $rr;  # MogileFS::ReplicationRequest
     while (1) {
         $rr = rr_upgrade($polobj->replicate_to(
                                                fid       => $fidid,
                                                on_devs   => \@on_devs_tellpol, # all device objects fid is on, dead or otherwise
-                                               all_devs  => $devs,
+                                               all_devs  => $dest_devs,
                                                failed    => \%dest_failed,
                                                min       => $cls->mindevcount,
                                                ));
